/*
 *    snake - a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999-2000 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* version string */
#define VERSION "0.7"

#include "snake.h"
#include <stdlib.h>		/* getenv, ... */
#include <time.h>		/* time (for srand) */
#include <sys/time.h>		/* itimer ... */
#include <glob.h>		/* glob_t etc. */
#include <string.h>		/* strcat, strcmp, ... */
#include <dlfcn.h>		/* dlopen, dlclose, ... */
#include <getopt.h>		/* getopt_long */
#include <signal.h>		/* signal() */
#include <setjmp.h>		/* setjmp, longjmp */
#include <signal.h>		/* signal */

/* extern functions */
extern int init_graphics (void);
extern void move_snakes (void);
extern void init_fruit (void);
extern int parse_input (void);
extern void draw_motion (void);
extern void print_delay (void);
extern void open_config (void);
extern void close_config (void);
extern void init_modules (void);
extern void draw_arena (void);
#ifdef ENABLE_MESSAGES
extern void expire_messages (void);
#endif

/* private functions */
static void snake_play (void);
static void parse_cmdline (int, char **);
static void usage (char *);
static void sighandler (int sig);
#ifdef PLAYERTIMEOUT
static void player_timeout (int sig);
#endif

/* public function */
int register_player (snake * s);
void finish (void);

/* public data */
snake *zoo[MAXPLAYERS];		/* snakes pointer */
int numplayers = 0;		/* number of players */
int delay = 40;			/* game delay in milliseconds */
char *rootdir;			/* installation root directory */
unsigned long int turns = 0;	/* game turns */
WINDOW *arenawin;


int
main (int argc, char **argv)
{

  /* parse command line */
  parse_cmdline (argc, argv);

  /* get installation root directory (default to ".") */
  if (!rootdir)
    rootdir = getenv ("SNAKEROOT");
  if (!rootdir)
    rootdir = ".";

  /* initialize the graphic or exit */
  if (!init_graphics ())
    {
      fprintf (stderr, "can't initialize graphic\n");
      finish ();
    }

  /* set signal handlers */
  signal (SIGINT, sighandler);

#ifdef PLAYERTIMEOUT
  signal (SIGVTALRM, player_timeout);	/* players timeout */
  signal (SIGALRM, player_timeout);	/* players timeout */
#endif

  /* init the arena */
  if (!arenainit (ARENA_HEIGHT, ARENA_WIDTH))
    {
      fprintf (stderr, "can't initialize arena\n");
      finish ();
    }

  /* report root directory value */
  fprintf (stderr, "SNAKEROOT=%s # set it yourself if needed\n", rootdir);

  /* initialize the random seed */
  srand (time (NULL));

  /* initialize the snakes */
  init_modules ();

  /* parse configuration file */
  parse_config ();

  /* initialize the fruit */
  init_fruit ();

  /* update the screen */
  updatescreen ();

  /* start ! */
#ifdef ENABLE_MESSAGES
  message ("ready steady go !");
#endif

  for (;;)
    {
      /* parse keyboard input */
      if (!parse_input ())
	break;

      /* give players a turn to play */
      snake_play ();

      /* move the snakes */
      move_snakes ();

      /* draw the arena */
      draw_arena ();

      /* show activity */
      draw_motion ();

      /* update the screen with the new data */
      updatescreen ();

      /* sleep for delay milliseconds (could use setitimer) */
      //for (i=0; i<delay*4000; i++);
      napms (delay);

      /* increment bogus time counter */
      turns++;

#ifdef ENABLE_MESSAGES
      /* expire messages */
      expire_messages ();
#endif

    }

  finish ();
  exit (0);			/* finish will exit itself */
}

void
finish ()
{
  int i;

  /* reset the screen */
  endwin ();

  printf ("\n");

  /* print scores */
  for (i = 0; zoo[i]; i++)
    {
      printf ("snake %i: score(%li), length(%i)\n",
	      i, zoo[i]->score, zoo[i]->length);
    }

  exit (0);
}


#ifdef PLAYERTIMEOUT
jmp_buf playloop;
#endif

static void
snake_play (void)
{
  int i;
#ifdef PLAYERTIMEOUT
  static struct itimerval playtimer = {
    {0, PLAYERTIMEOUT},		/* it_interval.{tv_sec,tv_usec} */
    {0, PLAYERTIMEOUT},		/* it_value.{tv_sec,tv_usec} */
  };
#endif


  for (i = 0; zoo[i]; i++)
    {

#ifdef PLAYERTIMEOUT
      if (setjmp (playloop))
	{

	  /* reset timer */
	  setitimer (TIMERMODE, NULL, NULL);

#ifdef ENABLE_MESSAGES
	  message ("!! snake %i thinks too much !!", i);
#endif

	  continue;
	}
#endif

      if (zoo[i]->died)
	continue;

      /* rewind typein queue */
      rewindtypein ();

#ifdef PLAYERTIMEOUT
      /* start timer */
      setitimer (TIMERMODE, &playtimer, NULL);
#endif

      /* call player function */
      (*zoo[i]->play) (zoo[i]);

#ifdef PLAYERTIMEOUT
      /* remove timer */
      setitimer (TIMERMODE, NULL, NULL);
#endif

    }
}

#ifdef PLAYERTIMEOUT
static void
player_timeout (int sig)
{
  /* reset signal handler (linux needs this, bsd doesn't) */
  signal (sig, player_timeout);
  longjmp (playloop, 1);
}
#endif


/* player modules must call this function at startup */
int
register_player (snake * s)
{
  /* don't register more then MAXPLAYERS snakes ! */
  /* gotta remove this limitation            */
  if (numplayers >= MAXPLAYERS - 1)
    {
      fprintf (stderr, "maximum count of players reached\n");
      return 0;
    }

  /* register new player */
  zoo[numplayers] = s;

  /* set numplayers flag */
  zoo[numplayers]->playernum = numplayers++;

  fprintf (stderr, "player %d registered\n", numplayers);

  return numplayers - 1;
}

/* get opts ... */
static void
parse_cmdline (int argc, char **argv)
{
  static struct option opts[] = {
    {"delay", 1, 0, 'd'},
    {"help", 0, 0, 'h'},
    {"version", 0, 0, 'v'},
    {"rootdir", 1, 0, 'r'},
    {0, 0, 0, 0}
  };
  int opts_idx;
  int c;

  while (1)
    {

      /* scan next option argument */
      c = getopt_long (argc, argv, "d:r:hv", opts, &opts_idx);

      /* end of options */
      if (c == -1)
	break;

      switch (c)
	{

	case 0:
	  fprintf (stderr, "bug in getopt?\n");
	  break;

	case 'd':
	  delay = atoi (optarg);
	  if (delay > MAXDELAY)
	    delay = MAXDELAY;
	  else if (delay < MINDELAY)
	    delay = MINDELAY;
	  break;

	case 'r':
	  rootdir = optarg;
	  break;

	case 'h':
	case '?':
	  usage (argv[0]);
	  break;		/* usage will exit anyway */

	case 'v':
	  printf ("snake %s\n", VERSION);
	  exit (0);

	default:
	  fprintf (stderr, "getopt_long returned %c (?)\n", c);

	}			/* end switch */

    }				/* end while */
}

static void
usage (char *me)
{
  fprintf (stderr, "Usage: %s [OPTIONS]\n", me);
  fprintf (stderr, "\n");
  fprintf (stderr, "  -d, --delay=time     set delay in millisecs\n");
  fprintf (stderr, "  -r, --rootdir=path   set SNAKEROOT\n");
  fprintf (stderr, "  -h, --help           show this screen and exit\n");
  fprintf (stderr,
	   "  -v, --version        output version information and exit\n");
  exit (1);
}

static void
sighandler (int sig)
{
/*
	signal(sig, sighandler);
#ifdef ENABLE_MESSAGES
	message("press 'q' to quit");
#endif
*/
  finish ();
}
