/*
 *    snake, a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999-2000 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "snake.h"
#include <stdlib.h>		/* malloc and friends */

/* extern data */
extern snake *zoo[];

/* private functions */
static void move_head (snake *);
static void eat_tail (snake *);
static void move_single_snake (snake *);
static int snake_crash (snakebody * s1, snakebody * s2);
static void eat_exsnake (snake * s);

/* public functions */
void move_snakes (void);

/* move each snake */
void
move_snakes ()
{
  int i;

  for (i = 0; zoo[i]; i++)
    {
      if (zoo[i]->died)
	continue;
      move_single_snake (zoo[i]);
    }

}

/* move head, eat tail */
void
move_single_snake (snake * s)
{
  /* move_head if not stopped */
  if (s->dir != STOP)
    move_head (s);

  /* finish here if snake died */
  if (s->died)
    return;

  /* eat the tail if not growing */
  if (!s->grow)
    eat_tail (s);
  else
    s->grow--;
}


/* move the snake's head */
static void
move_head (snake * s)
{
  snakebody *newhead;
  sprite *sp;

  /* allocate memory for new chain */
  newhead = (snakebody *) safe_malloc (sizeof (snakebody));

  /* set sprite type */
  newhead->type = SNAKE;

  /* set self reference */
  newhead->self = s;

  /* initialize newhead coordinates to old head's */
  newhead->cell.x = s->head->cell.x;
  newhead->cell.y = s->head->cell.y;

  /* move head one step in the desired direction */
  move_cell (&(newhead->cell), s->dir, 1);

  /* get sprite below the new head */
  sp = ARENAGET (newhead->cell.y, newhead->cell.x);

  /* check if snake have hitten something */
  if (IS_SNAKE (sp))
    {
      if (snake_crash (newhead, (snakebody *) sp))
	return;
    }

  /* check if snake have eaten something */
  else if (IS_FRUIT (sp))
    {				/* eat a fruit */

      int fruitval = FRUITVAL ((fruit *) sp);

      s->grow += fruitval;

      /* magic fruit (0 valued) */
      if (!fruitval)
	{
	  s->score += 10;
	  snake_trim (s);
#ifdef ENABLE_MESSAGES
	  message ("ZaC!");
#endif
	}
      else
	{
	  s->score += fruitval;
#ifdef ENABLE_MESSAGES
	  message ("Yum");
#endif
	}

      init_fruit ();

    }

  else if (IS_EXSNAKE (sp))
    {
      /* eat the snake's rest */
      eat_exsnake (s);
    }

  /* set the new snake's head */
  s->head->prev = newhead;
  newhead->next = s->head;
  newhead->prev = NULL;
  s->head = newhead;

  /* print the new head */
  /* do it (for now) to avoid transparency of 1-lenght snakes */
  ARENAPUT (newhead->cell.y, newhead->cell.x, (sprite *) newhead);

  /* increment the new snake's length */
  s->length++;

}


/* eat snake's tail */
static void
eat_tail (snake * s)
{
  /* chain pointer */
  snakebody *ptr;

  /* delete the old tail before its coordinates change */
  /* unless there is something else below it */
  /* this shouldn't be done by this module */
  ptr = (snakebody *) ARENAGET (s->tail->cell.y, s->tail->cell.x);
  if (ptr && ptr->self == s && ptr != ptr->self->head)
    ARENAPUT (s->tail->cell.y, s->tail->cell.x, NULL);

  /* find the chain before the tail */
  ptr = s->tail->prev;

  /* make it the new tail pointer */
  s->tail = ptr;

  /* free memory of the old tail */
  free (s->tail->next);

  /* set the new tail pointer to NULL */
  s->tail->next = NULL;

  /* decrement snake's lenght */
  s->length--;

  /* if snake disappeared, finish it */
  if (s->length == 0)
    snake_kill (s);
}


int
snake_crash (snakebody * s1, snakebody * s2)
{
#ifdef ENABLE_MESSAGES
  message ("snake_crash triggered");
#endif

  /* s2 is neither the tail, nor the head */
  if (s2 != s2->self->tail && s2 != s2->self->head)
    {
#ifdef ENABLE_MESSAGES
      message ("_cRaSh_");
#endif
      snake_kill (s1->self);
      return 1;
    }


  /* s2 is the tail ... */
  if (s2 == s2->self->tail)
    {

#ifdef ENABLE_MESSAGES
      message ("Fiuuuh ...");
#endif

      /* ... of an already moved snake */
      if (s2->self->playernum < s1->self->playernum)
	{
#ifdef ENABLE_MESSAGES
	  message ("_cRaSh_");
#endif
	  snake_kill (s1->self);
	}

      /* ... of a yet to move (but growing) snake */
      else if (s2->self->grow)
	{
#ifdef ENABLE_MESSAGES
	  message ("Nooo :(");
#endif
	  snake_kill (s1->self);
	}

    }

  /* s2 is the head ... */
  if (s2 == s2->self->head)
    {


      /* ... of an already moved snake */
      if (s2->self->playernum < s1->self->playernum ||
	  IS_OPPOSITE (s1->self->dir, s2->self->dir))
	{

#ifdef ENABLE_MESSAGES
	  message ("caboOOomm");
#endif

	  /* kill both snakes */
	  snake_kill (s1->self);
	  snake_kill (s2->self);
	}

      /* ... of a yet to move snake */
      else if (s2->self->length > 1)
	{
#ifdef ENABLE_MESSAGES
	  message ("_cRaSh_");
#endif
	  snake_kill (s1->self);
	}

      else
	{
#ifdef ENABLE_MESSAGES
	  message ("Swiiish ...");
#endif
	}

    }

  if (s1->self->died)
    return 1;

  return 0;
}

void
eat_exsnake (snake * s)
{
/*	static char blob[4] = { '.', 'o', 'O', 'o' }; */
  s->score++;

/*
#ifdef ENABLE_MESSAGES
		message("%c", blob[s->score%4]);
#endif
*/
}
